#pragma once
#include "Types.hpp"
#include <map>

class M3
{
public:
	// references and records //
	struct Record;					// contains a record of section type, address, size and flags
	struct Reference;				// references a record
	struct AnimationReference;		// references an animation
	struct AnimationIndex;			// contains 2 animation indices

	// sections //
	typedef char CHAR;		// types
	typedef uint8 U8;
	typedef uint16 U16;
	typedef int32 I32;
	typedef uint32 U32;
	typedef float REAL;
	typedef Vector2f VEC2;
	typedef Vector3f VEC3;
	typedef Quaternion4f VEC4, QUAT;
	typedef Matrix16f IREF;
	typedef Bounds7f BNDS;
	struct Vertex24;
	struct Vertex28;
	struct MD33;	// file header
	struct MODL20;	// model
	struct MODL23;	
	struct SEQS; 	// sequences
	struct STS;
	struct STC;
	struct STG;
	struct SD;		// sequence data
	struct SDEV;	
	struct SD3V;
	struct SDR3;
	struct SD4Q;
	struct SDMB;
	struct EVNT;
	struct BONE;	// bones
	struct ATT;
	struct DIV;		// mesh
	struct REGN;
	struct BAT;
	struct MSEC;
	struct MATM;	// materials
	struct MAT;
	struct LAYR;
	struct CAM;		// rendering
	struct PROG;
	struct LITE;
	struct SHBX;
	struct DIS;
	struct CMP;
	struct TER;
	struct VOL;
	struct CREP;
	struct PAR;
	struct PARC;
	struct RIB;
	struct FOR;
	struct WRP;
	struct PHRB;	// others
	struct IKJT;
	struct PATU;
	struct TRGD;
	struct SSGS;
	struct ATVL;
	struct BBSC;
	struct TMD;
	struct SSGS;

	// type-parser map type //
	typedef std::map<uint32, void(M3::*)(const Record&)> ParserMap;
	static ParserMap Parsers;
	static ParserMap Initialize();

	// type-parser callbacks //
	virtual void ParseCHAR(const Record& Record);
	virtual void ParseU8(const Record& Record);
	virtual void ParseU16(const Record& Record);
	virtual void ParseI32(const Record& Record);
	virtual void ParseU32(const Record& Record);
	virtual void ParseVEC2(const Record& Record);
	virtual void ParseVEC3(const Record& Record);
	virtual void ParseREAL(const Record& Record);
	virtual void ParseQUAT(const Record& Record);
	virtual void ParseIREF(const Record& Record);
	virtual void ParseBNDS(const Record& Record);
	virtual void ParseMD33(const Record& Record);
	virtual void ParseMODL(const Record& Record);
	virtual void ParseSEQS(const Record& Record);
	virtual void ParseSTG(const Record& Record);
	virtual void ParseSTC(const Record& Record);
	virtual void ParseSTS(const Record& Record);
	virtual void ParseSDEV(const Record& Record);
	virtual void ParseSD3V(const Record& Record);
	virtual void ParseSDR3(const Record& Record);
	virtual void ParseSD4Q(const Record& Record);
	virtual void ParseSDMB(const Record& Record);
	virtual void ParseDIV(const Record& Record);
	virtual void ParseREGN(const Record& Record);
	virtual void ParseBAT(const Record& Record);
	virtual void ParseMSEC(const Record& Record);
	virtual void ParseBONE(const Record& Record);
	virtual void ParseATT(const Record& Record);
	virtual void ParseEVNT(const Record& Record);
	virtual void ParseMATM(const Record& Record);
	virtual void ParseMAT(const Record& Record);
	virtual void ParseLAYR(const Record& Record);
	
	// public functions //
	static void ConvertM3toHTML(std::istream& In, std::ostream& Out);
	
	// members //
	std::istream& In;
	std::ostream& Out;
	Record* ReferenceTable;

	// methods //
	M3(std::istream& In, std::ostream& Out) : In(In), Out(Out) {ReferenceTable = NULL;}
	virtual ~M3();
	virtual void ConvertM3toHTML();
	virtual void Parse(const Record& Record);
	virtual void WriteRecordHTML(uint32 Index, const Record& Record);
	virtual void WriteReferenceHTML(const Reference& Reference, const char* Name = NULL);
	virtual void WriteReferenceHTML(const AnimationReference& Reference);
	virtual void WriteSectionStartHTML(const Record& Record);
	virtual void WriteSectionEndHTML();
	virtual void WriteHexDumpHTML(uint8 Unknown, const char* Name = NULL);
	virtual void WriteHexDumpHTML(int16 Unknown, const char* Name = NULL);
	virtual void WriteHexDumpHTML(uint16 Unknown, const char* Name = NULL);
	virtual void WriteHexDumpHTML(int32 Unknown, const char* Name = NULL);
	virtual void WriteHexDumpHTML(uint32 Unknown, const char* Name = NULL);
	virtual void WriteHexDumpHTML(float Unknown, const char* Name = NULL);
	virtual void WriteHexDumpHTML(const uint8* Unknown, uint Size, const char* Name = NULL);
	virtual void WriteHexDumpHTML(const int16* Unknown, uint Size, const char* Name = NULL);
	virtual void WriteHexDumpHTML(const uint16* Unknown, uint Size, const char* Name = NULL);
	virtual void WriteHexDumpHTML(const int32* Unknown, uint Size, const char* Name = NULL);
	virtual void WriteHexDumpHTML(const uint32* Unknown, uint Size, const char* Name = NULL);
	virtual void WriteHexDumpHTML(const float* Unknown, uint Size, const char* Name = NULL);
	virtual void WriteVerticesHTML(const Reference& Reference, uint32 Flags);
	friend std::ostream& operator << (std::ostream& Out, const VEC2& Vector);
	friend std::ostream& operator << (std::ostream& Out, const VEC3& Vector);
	friend std::ostream& operator << (std::ostream& Out, const QUAT& Quaternion);
	friend std::ostream& operator << (std::ostream& Out, const BNDS& Bounds);
	friend std::ostream& operator << (std::ostream& Out, const IREF& Matrix);
	friend std::ostream& operator << (std::ostream& Out, const Vertex24& Vertex);
	friend std::ostream& operator << (std::ostream& Out, const Vertex28& Vertex);

	// template functions //
	template <typename ArrayType>
	void ParseArray(const Record& Record)
	{
		ArrayType* Array = new ArrayType [Record.Size];

		In.read((char*)Array, Record.Size * sizeof(ArrayType));
		Out << "<div>{";
		for(uint32 i = 0; i < Record.Size; ++i){
			Out << Array[i] << ", ";
		}
		Out << "}</div>";

		delete [] Array;
	}
	template <typename SectionType>
	void ParseSection(const Record& Record)
	{
		SectionType Section;

		In.read((char*)&Section, sizeof(SectionType));
		Out << "<div>\n";
		Section.WriteHTML(*this);
		Out << "</div>\n";
	}
	template <typename SectionType>
	void ParseSections(const Record& Record)
	{
		SectionType* Sections = new SectionType [Record.Size];

		In.read((char*)Sections, Record.Size * sizeof(SectionType));
		for(uint32 i = 0; i < Record.Size; ++i){
			Out << "<div>\n";
			Sections[i].WriteHTML(*this);
			Out << "</div>\n";
		}

		delete [] Sections;
	}
	template <typename Type>
	void WriteHex(std::ostream& Out, Type Value){
		uint32 Mask;

		Mask = Value & 0xFF;
		if(Mask < 0x10){
			Out << 0;
		}
		Out << Mask << " ";
		if(sizeof(Value) == 1){
			return;
		}

		Mask = ((uint32&)Value & 0x0000FF00) >> 8;
		if(Mask < 0x10){
			Out << 0;
		}
		Out << Mask << " ";
		if(sizeof(Value) == 2){
			return;
		}

		Mask = (Value & 0x00FF0000) >> 16;
		if(Mask < 0x10){
			Out << 0;
		}
		Out << Mask << " ";
		Mask = ((uint32&)Value >> 24);
		if(Mask < 0x10){
			Out << 0;
		}
		Out << Mask << " ";
	}
};

struct M3::Record
{
	tag32 Type;
	uint32 Address;
	uint32 Size;
	uint32 Flags;
};

struct M3::Reference
{
	uint32 Size;
	uint32 Index;
};

struct M3::AnimationReference
{
	uint16 Flags; 			//usually 1
	uint16 AnimationFlags; 	//6 indicates animation data present
	uint32 AnimationID;		//a unique uint32 value referenced in STC.animid and STS.animid
};

struct AnimationIndex
{
	uint16 FrameIndex;			// frame index of sequence data block
	uint16 SequenceDataIndex;	// index into sequence data array
};

struct M3::MD33
:	public M3::Record
{
	uint32 RootIndex;

	void WriteHTML(M3& ParserM3) const;
};

struct M3::MODL20
{
	Reference MaxFilenameCHAR;
	uint32 Version;
	Reference SequencesSEQS;
	Reference SequencesSTC;
	Reference SequencesSTG;
	uint32 Unknown0 [3];
	Reference SequencesSTS;
	Reference BonesBONE;
	uint32 Unknown1;
	uint32 Flags;
	Reference VerticesU8;
	Reference GeometryDIV;
	Reference BoneLookupU16;
	BNDS Bounds;
	uint32 Unknown2 [13];
	Reference AttachmentATT;
	Reference AttachmentLookupU16;
	Reference Lights;
	Reference Cameras;
	Reference A_U16;
	Reference MaterialGroupsMATM;
	Reference MaterialMAT;
	Reference DIS;
	Reference CMP;
	Reference TER; 
	Reference VOL;
	uint32 Unknown3 [2];
	Reference CREP;
	Reference PAR;
	Reference PARC;
	Reference RIB;
	Reference PROJ;
	Reference FOR;
	Reference WRP;
	uint32 Unknown4 [2];
	Reference PHRB;
	uint32 Unknown5 [6];
	Reference IKJT;
	uint32 Unknown6 [2];
	Reference PATU;
	Reference TRGD;
	Reference BonesIREF;
	Reference B_I32;
	Matrix16f Matrix;
	BNDS Bounds1;
	Reference SSGS;
	Reference ATVL;
	Reference BBSC;
	Reference TDM;
	uint32 Unknown7 [3];

	void WriteHTML(M3& ParserM3) const;
};

struct M3::MODL23
{
	Reference MaxFilenameCHAR;
	uint32 Version;
	Reference SequencesSEQS;
	Reference SequencesSTC;
	Reference SequencesSTG;
	uint32 Unknown0 [3];
	Reference SequencesSTS;
	Reference BonesBONE;
	uint32 Unknown1;
	uint32 Flags;
	Reference VerticesU8;
	Reference GeometryDIV;
	Reference BoneLookupU16;
	BNDS Bounds;
	uint32 Unknown2 [13];
	Reference AttachmentATT;
	Reference AttachmentLookupU16;
	Reference Lights;
	Reference SHBX;
	Reference Cameras;
	Reference A_U16;
	Reference MaterialGroupsMATM;
	Reference MaterialMAT;
	Reference DIS;
	Reference CMP;
	Reference TER; 
	Reference VOL;
	uint32 Unknown3 [2];
	Reference CREP;
	Reference PAR;
	Reference PARC;
	Reference RIB;
	Reference PROJ;
	Reference FOR;
	Reference WRP;
	uint32 Unknown4 [2];
	Reference PHRB;
	uint32 Unknown5 [6];
	Reference IKJT;
	uint32 Unknown6 [2];
	Reference PATU;
	Reference TRGD;
	Reference BonesIREF;
	Reference B_I32;
	Matrix16f Matrix;
	BNDS Bounds1;
	Reference SSGS;
	Reference ATVL;
	Reference C_U16;
	Reference D_U16;
	Reference BBSC;
	Reference TDM;
	uint32 Unknown7 [3];

	void WriteHTML(M3& ParserM3) const;
};

struct M3::DIV
{
	Reference FacesU16;
	Reference RegionsREGN;
	Reference MaterialsBAT;
	Reference MSEC;
	uint32 unk;

	void WriteHTML(M3& Parser) const;
};

struct M3::REGN
{
	uint32 d1; 				// always 0?
	uint16 VertexStart;		
	uint16 VertexCount;		
	uint32 IndexStart;		
	uint32 IndexCount;
	uint16 MaxBoneSize;
	uint16 BoneStart;
	uint16 BoneCount;
	uint16 s1 [3];

	void WriteHTML(M3& Parser) const;
};

struct M3::BAT
{
	uint32 d1; 				// always 0?
	uint16 RegionIndex; 	// associated REGN
	uint16 s1[2];
	uint16 MaterialIndex; 	// index into MATM table
	uint16 s2; 				// usually -1

	void WriteHTML(M3& Parser) const;
};

struct M3::MSEC
{
	uint32 d1; 							// always 0?
	AnimationReference BoundsAnimation; // Bounding box animation ref in STC
	Vector3f ext1 [2];					// initial bounds?
	uint32 d2[8];

	void WriteHTML(M3& Parser) const;
};

struct M3::SEQS
{
	uint32 d1 [2];
	Reference SequenceNameCHAR;
	uint32 d2;
	uint32 TimeLength;
	uint32 d3;
	uint32 Flags;
	uint32 d4 [5];
	Bounds7f Bounds;
	uint32 d5 [2];

	void WriteHTML(M3& Parser) const;
};

struct M3::STC
{
    Reference SequenceNameCHAR;
    uint32 d1;
    uint16 SequenceIndex;		// points to animation in SEQS chunk, twice for some reason
    uint16 SequenceIndex1;
    Reference AnimationIDsU32;	
    Reference AnimationLookupU32;
    uint32 d2;
	union {
    	Reference SequenceData[13];
		struct {
			Reference TranslationSD3V;
			Reference RotationSD4Q;
			Reference ScaleSDR3;
			Reference FlagsSDFG;
			Reference BoundsSDMB;
		};
	};

	void WriteHTML(M3& Parser) const;
};

struct M3::STS
{
    Reference AnimationIDsU32;
    uint32 d1[3];
    uint16 s1;
    uint16 s2;

	void WriteHTML(M3& ParserM3) const;
};

struct M3::STG
{
    Reference SequenceNameCHAR;
    Reference SequenceIDU32;

	void WriteHTML(M3& ParserM3) const;
};

struct M3::SD
{
	Reference TimelineI32;
	uint32 Flags;
	uint32 TimeLength;

	void WriteHTML(M3& ParserM3) const;
};

struct M3::SDEV
:	public M3::SD
{
	Reference DataEVNT;

	void WriteHTML(M3& ParserM3) const;
};

struct M3::SD3V
:	public M3::SD
{
	Reference DataVEC3;

	void WriteHTML(M3& ParserM3) const;
};

struct M3::SDR3
:	public M3::SD
{
	Reference DataREAL;

	void WriteHTML(M3& ParserM3) const;
};

struct M3::SD4Q
:	public M3::SD
{
	Reference DataQUAT;

	void WriteHTML(M3& ParserM3) const;
};

struct M3::SDMB
:	public M3::SD
{
	Reference DataBNDS;

	void WriteHTML(M3& ParserM3) const;
};

struct M3::BONE
{
	uint32 d1; 	// Index into something
	Reference BoneNameCHAR;
	uint32 Flags; // 0x800 = used in skinning, always 0x200?
	int16 ParentIndex;
	uint16 s1;
	AnimationReference TranslationAnimation;
	Vector3f Translation; // point of rotation (i.e. joint)
	uint32 d2[4];
	AnimationReference RotationAnimation;
	Quaternion4f Rotation; // rotation at rest
	uint32 d3[3];
	float f1;
	uint32 d4;
	AnimationReference ScaleAnimation;
	Vector3f Scale; // scale at rest
	Vector3f v1;
	uint32 d5[6]; //appears to have another animation ref at 0x8C

	void WriteHTML(M3& ParserM3) const;
};

struct M3::MATM
{
	uint32 Size;			// number of materials
	uint32 MaterialIndex;	// index into MAT

	void WriteHTML(M3& ParserM3) const;
};

struct M3::MAT
{
    Reference MaterialNameCHAR;
    uint32 d1[8];
    float x, y;
	union {
    	Reference Layers [13];
		struct {
			Reference DiffuseLAYR;
			Reference DecalLAYR;
			Reference SpecularLAYR;
			Reference EmissiveLAYR;
			Reference Layer4;
			Reference Layer5;
			Reference Layer6;
			Reference Layer7;
			Reference Layer8;
			Reference NormalLAYR;
			Reference Layer10;
			Reference Layer11;
			Reference Layer12;
		};
	};
    uint32 d2[15];

	void WriteHTML(M3& ParserM3) const;
};

struct M3::LAYR
{
    uint32 d1;
    Reference LayerNameCHAR;
    AnimationReference ar1;
    uint32 d2[6]; //d2[5] determines team colour
    AnimationReference ar2;
    float f1[2];
    uint32 d3;
    AnimationReference ar3;
    uint32 d4[4];
    int32 d5;
    uint32 d6[5];
    AnimationReference ar4;
    uint32 d7[3];
    AnimationReference ar5;
    uint32 d8[5];
    AnimationReference ar6;
    uint32 d9[2];
    AnimationReference ar7;
    float f2[2];
    uint32 d10[3];
    AnimationReference ar8;
    float f3;
    uint32 d11;
    float f4;
    uint32 d12[4];
    AnimationReference ar9;
    float f5[4];
    uint32 d13;
    AnimationReference ar10;
    uint32 d14[3];
    AnimationReference ar11;
    float f6[2];
    uint32 d15;
    AnimationReference ar12;
    float f7[2];
    uint32 d16;
    int32 d17;
    uint32 d18;
    float f8;
    uint32 d19;
    float f9[2];

	void WriteHTML(M3& ParserM3) const;
};

struct M3::CAM
{
	int32 dl;
	Reference CameraNameCHAR;
	uint16 flags1;
	uint16 flags2;

	void WriteHTML(M3& ParserM3) const;
};

struct M3::EVNT
{
	Reference EventNameCHAR;
	int16 s1[4];
	IREF Matrix;
	int32 d1 [4];

	void WriteHTML(M3& ParserM3) const;
};

struct M3::ATT
{
	int32 d1;
	Reference AttachmentNameCHAR;
	int32 BoneIndex;

	void WriteHTML(M3& ParserM3) const;
};

struct M3::Vertex24
{
	Vector3f Position;
	uint8 BoneWeights[4];
	uint8 BoneIndices[4];
	uint8 Normal[4];
	int16 UV[2];
	uint8 Tangent[4];
};

struct M3::Vertex28
{
	Vector3f Position;
	uint8 BoneWeights[4];
	uint8 BoneIndices[4];
	uint8 Normal[4];
	int16 UV[2];
	int16 s1[2];
	uint8 Tangent[4];
};